package htsjdk.beta.plugin;

// Note: This is a tagging interface currently used as a type-bound for codec/encoder/decoder
// record type params, and used to tag SAMRecord, BAMRecord, VariantContext, and ReferenceSequence.
// It will be more useful once version upgrading is implemented.

/**
 * Tagging interface used as a type-bound for codec/encoder/decoder record type params.
 */
public interface HtsRecord {}
