/**
 * ****************************************************************************
 * Copyright 2013 EMBL-EBI
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ****************************************************************************
 */
package htsjdk.samtools.cram.encoding.core.huffmanUtils;

import java.util.Map;

/**
 * A leaf node in a {@code HuffmanTree}.
 * @param <T> type of the symbols in the alphabet being huffman-encoded
 */
final class HuffmanLeaf<T> extends HuffmanTree<T> {
    public final T symbol;

    /**
     * A symbol in the alphabet being encoded, and the corresponding frequency in the input stream.
     * @param symbol
     * @param frequency
     */
    public HuffmanLeaf(final T symbol, final int frequency) {
        super(frequency);
        this.symbol = symbol;
    }

    @Override
    public void getCodeWords(int codeWord, int codeWordLength, final Map<T, HuffmanBitCode<T>> symbolsToCodes) {
        symbolsToCodes.put(symbol, new HuffmanBitCode(symbol, codeWord, codeWordLength));
    }
}
